#ifndef _BRANCHING_H_
#define _BRANCHING_H_

#include<iostream>

template<typename Node>
inline bool node_finished(Node* node) {
	return node->lowerChild && node->upperChild;
}

template<typename Node>
inline bool node_unvisited(Node* node) {
	return !node->lowerChild && !node->upperChild;
}

template<typename Node>
inline bool lower_child_unvisited(Node* node) {
	return !node->lowerChild;
}

template<typename Node>
inline bool upper_child_unvisited(Node* node) {
	return !node->upperChild;
}

template<typename Node, class Traits>
inline bool node_further_explorable(Node* node, Traits& traits) {
	return node->feasible && !traits.tree.worse(node->bound);
}

template<typename Node, class Traits>
Node* go_upwards(Node* node, Traits& traits, const int& i) {
	Node* parentNode = node->parent;
	traits.solver[i].set_bounds(parentNode->variable, parentNode->lowerBound,
			parentNode->upperBound);
	return parentNode;
}

template<typename Node, class Traits>
Node* new_lower_child(Node* node, Traits& traits, const int& i) {
	++traits.tree.num_nodes;
//	traits.solver[i].set_bounds(node->variable, node->lowerBound,
//			floor(node->value));
	traits.solver[i].set_bounds(node->variable, node->lowerBound,
			node->lowerBound);
	Node* lowerChild = new Node;
	lowerChild->parent = node;
	node->lowerChild = lowerChild;
	lowerChild->depth = node->depth + 1;
	return lowerChild;
}

template<typename Node, class Traits>
Node* new_upper_child(Node* node, Traits& traits, const int& i) {
	++traits.tree.num_nodes;
//	traits.solver[i].set_bounds(node->variable, ceil(node->value),
//			node->upperBound);
	traits.solver[i].set_bounds(node->variable, node->upperBound,
			node->upperBound);
	Node* upperChild = new Node;
	upperChild->parent = node;
	node->upperChild = upperChild;
	upperChild->depth = node->depth + 1;
	return upperChild;
}

template<typename Node, class Traits>
Node* get_best_node(Node* node, Traits& traits, const int& i) {
	Node* traverse = node;
	Node* nextNode = traits.tree.get_next_node();
	if (traverse != traits.tree.root
			&& (!traits.tree.foundFeasibleSolution || nextNode == NULL))
		return go_upwards(traverse, traits, i);
	while (nextNode
			&& (node_finished(nextNode) || traits.tree.worse(nextNode->bound)))
		nextNode = traits.tree.get_next_node();
	while (traverse != nextNode && traverse != traits.tree.root)
		traverse = go_upwards(traverse, traits, i);
	if (traverse == nextNode || nextNode == NULL)
		return nextNode;
	traverse = nextNode;
	while (traverse->parent) {
		if (traverse == traverse->parent->lowerChild)
			traits.solver[i].set_bounds(traverse->parent->variable,
					traverse->parent->lowerBound,
					traverse->parent->lowerBound);
//					floor(traverse->parent->value));
		else
			traits.solver[i].set_bounds(traverse->parent->variable,
//					ceil(traverse->parent->value),
					traverse->parent->upperBound,
					traverse->parent->upperBound);
		traverse = traverse->parent;
	}
	return nextNode;
}

template<bool queue, typename Node, class Traits>
Node* examine_node(Node* currentNode, Traits& traits, const int& i) {
	if (!currentNode)
		currentNode = traits.tree.root;

	traits.tree.propagate_bound(currentNode);

	if (node_finished(currentNode)) {
		if (currentNode == traits.tree.root) {
			if (queue)
				return get_best_node(currentNode, traits, i);
			else
				return NULL;
		}
		return get_best_node(currentNode, traits, i);
	}

	if (node_unvisited(currentNode)) {
		traits.solver[i].solve();

		if (traits.solver[i].lp_infeasible()) {
			currentNode->feasible = false;
			traits.node_infeasible_notification(i, currentNode);
			return get_best_node(currentNode, traits, i);
		}

		if (traits.solver[i].lp_feasible()) {
			currentNode->feasible = true;
			currentNode->bound = traits.solver[i].objective();

			if (traits.tree.worse(currentNode->bound)) {
				traits.node_worse_notification(currentNode);
				return get_best_node(currentNode, traits, i);
			}

			if (traits.solver[i].lp_integral()) {
				const double qap_obj = traits.solver[i].QAP_objective();
				traits.node_integral_notification(currentNode, qap_obj);
				if (qap_obj
						<= currentNode->bound + traits.solver[i].eps_integrality
						|| traits.solver[i].get_candidate() == -1) {
					return get_best_node(currentNode, traits, i);
				}
			}

			if (queue)
				traits.tree.push_node(currentNode);
			currentNode->variable = traits.solver[i].get_candidate();
			currentNode->value = traits.solver[i].value_of(
					currentNode->variable);
			currentNode->lowerBound = traits.solver[i].lower_bound_of(
					currentNode->variable);
			currentNode->upperBound = traits.solver[i].upper_bound_of(
					currentNode->variable);
			//			currentNode->basis = 
			//				traits.basis_store.push( traits.solver[i].get_basis() );

			if (traits.lower_first(i))
				return new_lower_child(currentNode, traits, i);
			else
				return new_upper_child(currentNode, traits, i);

		}

	}

	//	traits.solver[i].load_basis( traits.basis_store.get_basis( currentNode->basis ) );

	if (lower_child_unvisited(currentNode)) {
		return new_lower_child(currentNode, traits, i);
	}

	if (upper_child_unvisited(currentNode)) {
		return new_upper_child(currentNode, traits, i);
	}

	return NULL;
}

template<bool queue, typename Node, class Traits>
Node* examine_current_node(Node* currentNode, Traits& traits, const int& i) {
	traits.solver[i].solve();
	currentNode->solved = true;

	if (traits.solver[i].fallback()) {
		currentNode->feasible = true;
		currentNode->bound = currentNode->parent->bound;
	} else if (traits.solver[i].lp_infeasible()) {
		currentNode->feasible = false;
		traits.node_infeasible_notification(i, currentNode);
		return get_best_node(currentNode, traits, i);
	} else if (traits.solver[i].lp_feasible()) {
		currentNode->feasible = true;
		currentNode->bound = traits.solver[i].objective();

		if (traits.tree.worse(currentNode->bound)) {
			traits.node_worse_notification(currentNode);
			return get_best_node(currentNode, traits, i);
		}

		if (traits.solver[i].lp_integral()) {
			const double qap_obj = traits.solver[i].QAP_objective();
			traits.node_integral_notification(currentNode, qap_obj);
			return get_best_node(currentNode, traits, i);
		}

	}
	if (queue)
		traits.tree.push_node(currentNode);
	currentNode->variable = traits.solver[i].get_candidate();
	currentNode->value = traits.solver[i].value_of(currentNode->variable);
	currentNode->lowerBound = traits.solver[i].lower_bound_of(
			currentNode->variable);
	currentNode->upperBound = traits.solver[i].upper_bound_of(
			currentNode->variable);

	return currentNode;
}

template<bool queue, typename Node, class Traits>
void solve_only(Node* currentNode, Traits& traits, const int& i) {
	traits.solver[i].solve();
	currentNode->solved = true;

	if (traits.solver[i].fallback()) {
		currentNode->feasible = true;
		currentNode->bound = currentNode->parent->bound;
	} else if (traits.solver[i].lp_infeasible()) {
		currentNode->feasible = false;
		traits.node_infeasible_notification(i, currentNode);
		return;
	} else if (traits.solver[i].lp_feasible()) {
		currentNode->feasible = true;
		currentNode->bound = traits.solver[i].objective();

		if (traits.tree.worse(currentNode->bound)) {
			traits.node_worse_notification(currentNode);
		}

		if (traits.solver[i].lp_integral()) {
			const double qap_obj = traits.solver[i].QAP_objective();
			traits.node_integral_notification(currentNode, qap_obj);
		}
	}
	traits.tree.push_node(currentNode);
	currentNode->variable = traits.solver[i].get_candidate();
	currentNode->value = traits.solver[i].value_of(currentNode->variable);
	currentNode->lowerBound = traits.solver[i].lower_bound_of(
			currentNode->variable);
	currentNode->upperBound = traits.solver[i].upper_bound_of(
			currentNode->variable);
}

template<bool queue, typename Node, class Traits>
Node* explore_all_children(Node* currentNode, Traits& traits, const int& i) {
	if (!currentNode)
		currentNode = traits.tree.root;

	if (node_finished(currentNode)) {
		return get_best_node(currentNode, traits, i);
	}

	if (node_unvisited(currentNode)) {
		if (!currentNode->solved) {
			return examine_current_node<queue>(currentNode, traits, i);
		}

		if (!node_further_explorable(currentNode, traits))
			return get_best_node(currentNode, traits, i);
		Node* lower = new_lower_child(currentNode, traits, i);
		solve_only<queue>(lower, traits, i);
		go_upwards(lower, traits, i);

		Node* upper = new_upper_child(currentNode, traits, i);
		solve_only<queue>(upper, traits, i);
		go_upwards(upper, traits, i);

		traits.tree.propagate_bound(currentNode);

//		if (node_further_explorable(lower, traits)
//				|| node_further_explorable(upper, traits))
//			return currentNode;
		return get_best_node(currentNode, traits, i);

	}

	return NULL;
}

template<class Traits>
bool optimal(Traits& traits){
	return traits.tree.root->bound == traits.tree.bestKnownIntegralSolution;
}

template<int num, typename Node>
bool not_all_null(Node* node[num]) {
	bool not_null = false;
	for (int i = 0; i < num; ++i)
		not_null |= (node[i] != NULL);
	return not_null;
}

template<int num, class Traits>
bool not_all_done(Traits& traits){
	bool not_done = false;
	for (int i = 0; i < num; ++i)
		not_done |= !traits.solver[i].done;
	return not_done;
}

template<int num_solvers, class Traits>
void branching(Traits& traits) {
	typedef typename Traits::Node Node;

	Node* currentNode[num_solvers];

	unsigned long iterations = 0;

	for (int i = 0; i < num_solvers; ++i)
		currentNode[i] = traits.tree.root;

	traits.produce_heuristic_solution(traits.tree.root);

	while (not_all_null<num_solvers>(currentNode) && not_all_done<num_solvers>(traits) && !optimal(traits)) {
		for (int i = 0; i < num_solvers; ++i) {
//			currentNode[i] = examine_node<(num_solvers > 1)>(currentNode[i],
//					traits, i);
			currentNode[i] = explore_all_children<true>(
					currentNode[i], traits, i);
			++iterations;
		}
		if (iterations == 1){
			std::cout << "Root relaxation: " << traits.tree.root->bound << endl
				  << "Heurisitc incumbent: " << traits.tree.bestKnownIntegralSolution << endl;
		}
		if (iterations % 10ul == 0ul) {
			std::cout << iterations << " iterations, ";
			const int depth = currentNode[0] ? currentNode[0]->depth : -1;
			std::cout << traits.tree.num_nodes
					<< " nodes in the tree, current depth: "
					<< depth;
			std::cout << ", global LP-bound: " << traits.tree.root->bound;
			if (traits.tree.foundFeasibleSolution) {
				std::cout << ", current best solution: "
						<< traits.tree.bestKnownIntegralSolution;
			} else {
				std::cout << ", no feasible solution yet";
			}
			std::cout << std::endl;
		}
	}

}

#endif
