#ifndef MOSEKSOLVER_H_
#define MOSEKSOLVER_H_

extern "C" {
#include "mosek.h"
}
#include <math.h>
#include <stdlib.h>
#include <vector>
#include "matutils.h"
#include <bitset>
#include <fstream>
#include <iostream>
#include "sdpsolver.h"

using namespace std;

class mosekSolver: public sdpSolver {

public:

	typedef int Basis;

	MSKenv_t env = NULL;
	MSKtask_t task = NULL;
	MSKrescodee res;
	bool feasible;
	MSKint64t *branch_constraints;
	MSKint64t *running_ids;
	MSKint64t *product_ids;
	MSKint64t z_id;
	MSKint64t values_in_lower_triangle;
	MSKint32t constraint_number;
	MSKint32t *cost_constraint;
	std::vector<MSKrealt*> cost_matrix;


	vector<vector<vector<long>>> fixationIDs11;
	vector<vector<vector<long>>> fixationIDs10;
	vector<vector<vector<long>>> fixationIDs01;
	vector<vector<vector<long>>> fixationIDs00;

	double *Y;

	MSKsolstae solsta;
	MSKprostae prosta;
	MSKstakeye *skc;
	MSKstakeye *skx;
	MSKstakeye *skn;
	MSKrealt *xc;
	MSKrealt *xx;
	MSKrealt *y;
	MSKrealt *slc;
	MSKrealt *suc;
	MSKrealt *slx;
	MSKrealt *sux;
	MSKrealt *snx;

	mosekSolver() :
			res(MSK_RES_OK), feasible(false), branch_constraints(
			NULL), running_ids(NULL), z_id(0), values_in_lower_triangle(0), constraint_number(
					0), cost_constraint(NULL), Y(NULL), solsta(
					MSK_SOL_STA_UNKNOWN), prosta(MSK_PRO_STA_UNKNOWN), skc(
			NULL), skx(NULL), skn(NULL), xc(
			NULL), xx(NULL), y(NULL), slc(NULL), suc(NULL), slx(NULL), sux(
			NULL), snx(NULL) {

	}
	;

	~mosekSolver() {
		if (task)
			MSK_deletetask(&task);
		if (env)
			MSK_deleteenv(&env);
	}

	void checkSanityOfSetup() {

		if (res == MSK_RES_OK) {
			res = MSK_putintparam(task, MSK_IPAR_NUM_THREADS, 1);
		}

		if (res != MSK_RES_OK) {
			std::cerr << "Unable to create the mosek environment: " << res
					<< std::endl;
			exit(1);
		}

	}

	void init_env() {
		res = MSK_makeenv(&env, NULL);

		if (res == MSK_RES_OK)
			res = MSK_maketask(env, 0, 0, &task);

		MSKint32t *dimSdpVars = (MSKint32t *) malloc(
				numCuts * sizeof(MSKint32t));
		for (int i = 0; i < numCuts; ++i) {
			dimSdpVars[i] = dim;
		}

		if (res == MSK_RES_OK)
			res = MSK_appendcons(task, numConstraints);

		if (res == MSK_RES_OK)
			res = MSK_appendbarvars(task, numCuts, dimSdpVars);

		if (res == MSK_RES_OK)
			res = MSK_appendvars(task, dim * numCuts);

		for (int i = 0; i < dim * numCuts && res == MSK_RES_OK; ++i) {
			res = MSK_putvarbound(task, i, MSK_BK_RA, -1.0, 1.0);
		}


		Y = (double*) malloc(dim * dim * sizeof(double));

		malloc_solution_container();
	}

	void add_max_variable() {
		int numvars;
		MSK_getnumvar(task, &numvars);
		z_id = (MSKint64t) numvars;
		if (res == MSK_RES_OK)
			res = MSK_appendvars(task, 1);
		if (res == MSK_RES_OK)
			res = MSK_putvarbound(task, z_id, MSK_BK_FR, 0.0, 0.0);
		cost_constraint = (MSKint32t *) malloc(numCuts * sizeof(MSKint32t));
	}

	void declareFixationVariables() {
		MSKint64t running_id = dim * numCuts;
		numfixvar = 4 * dimNoDummies * (dimNoDummies - 1) * numCuts
				+ 2 * dimNoDummies * numCuts;

		if (res == MSK_RES_OK)
			res = MSK_appendvars(task, numfixvar);

		for (int i = running_id;
				i < running_id + numfixvar && res == MSK_RES_OK; ++i) {
			res = MSK_putvarbound(task, i, MSK_BK_RA, 0, 1);
		}

		fixationIDs11.resize(numCuts);
		fixationIDs10.resize(numCuts);
		fixationIDs01.resize(numCuts);
		fixationIDs00.resize(numCuts);
		for (int b = 0; b < numCuts; ++b) {
			fixationIDs11[b].resize(dimNoDummies);
			fixationIDs10[b].resize(dimNoDummies);
			fixationIDs01[b].resize(dimNoDummies);
			fixationIDs00[b].resize(dimNoDummies);
			for (int i = 0; i < dimNoDummies; ++i) {
				fixationIDs11[b][i].resize(dimNoDummies);
				fixationIDs10[b][i].resize(dimNoDummies);
				fixationIDs01[b][i].resize(dimNoDummies);
				fixationIDs00[b][i].resize(dimNoDummies);
				for (int j = 0; j < dimNoDummies; ++j) {
					if (i == j) {
						fixationIDs11[b][i][i] = running_id++;
						fixationIDs00[b][i][i] = running_id++;
					} else {
						fixationIDs11[b][i][j] = running_id++;
						fixationIDs10[b][i][j] = running_id++;
						fixationIDs01[b][i][j] = running_id++;
						fixationIDs00[b][i][j] = running_id++;
					}
				}
			}
		}

	}

	void addProductVariableAndConstraints() {
		int dimension = dimNoDummies * (dimNoDummies + 1) / 2 + dimNoDummies;

		int numVarsSoFar;
		MSK_getnumbarvar(task, &numVarsSoFar);

		MSKint32t *dimSdpVars = (MSKint32t *) malloc(
				numCuts * sizeof(MSKint32t));
		for (int i = 0; i < numCuts; ++i) {
			dimSdpVars[i] = dimension;
		}
		if (res == MSK_RES_OK)
			res = MSK_appendbarvars(task, numCuts, dimSdpVars);

		int additionalConstraints = (dimNoDummies + 1)
				* (dimNoDummies + (dimNoDummies + 1) / 2 + dimNoDummies);
		additionalConstraints = additionalConstraints * numCuts;

		if (res == MSK_RES_OK)
			res = MSK_appendcons(task, additionalConstraints);

		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j <= i && res == MSK_RES_OK; ++j) {
				int indexI = i * (i + 1) / 2 + j;
				double value = 1.0;
				MSKint64t id;
				res = MSK_appendsparsesymmat(task, dimension, 1, &indexI,
						&indexI, &value, &id);
				for (int b = 0; b < numCuts; ++b) {
					if (res == MSK_RES_OK)
						res = MSK_putbaraij(task, constraint_number,
								numVarsSoFar + b, 1, &id, &value);
					if (res != MSK_RES_OK) {
						exit(1);
					}
					if (res == MSK_RES_OK)
						res = MSK_putconbound(task, constraint_number++,
								MSK_BK_FX, 1.0, 1.0);
				}
				for (int k = 0; k < dimNoDummies; ++k) {
					int jj = max(j, k);
					int kk = min(j, k);
					int indexJ = jj * (jj + 1) / 2 + kk;
					int indexII = max(indexI, indexJ);
					int indexJJ = min(indexI, indexJ);
					if (res == MSK_RES_OK)
						res = MSK_appendsparsesymmat(task, dimension, 1,
								&indexII, &indexJJ, &value, &id);
					if (res != MSK_RES_OK) {
						exit(1);
					}
					double weightPlus = 1.0;
					double weightMinus = -1.0;
					int ii = max(i, k);
					kk = min(i, k);
					int index = IJtoIndex(ii, kk);
					assert(index < values_in_lower_triangle);
					MSKint64t id_orig = running_ids[index];
					for (int b = 0; b < numCuts; ++b) {
						if (res == MSK_RES_OK)
							res = MSK_putbaraij(task, constraint_number,
									numVarsSoFar + b, 1, &id, &weightPlus);
						if (res != MSK_RES_OK) {
							exit(1);
						}
						if (res == MSK_RES_OK)
							res = MSK_putbaraij(task, constraint_number, b, 1,
									&id_orig, &weightMinus);
						if (res != MSK_RES_OK) {
							exit(1);
						}
						if (res == MSK_RES_OK)
							res = MSK_putconbound(task, constraint_number++,
									MSK_BK_FX, 0.0, 0.0);
						if (res != MSK_RES_OK) {
							exit(1);
						}
					}
				}
			}
		}
	}

	void declare_constraints() {
		const double weight1 = 1.0;
		const double weight5 = 0.5;


		constraint_number = 0;
		values_in_lower_triangle = dim * (dim + 1) / 2;
		MSKint64t nonzeroes = 1;
		running_ids = (MSKint64t *) malloc(
				values_in_lower_triangle * sizeof(MSKint64t));
		int runner = 0;
		for (int i = 0; i < dim; ++i) {
			for (int j = 0; j <= i; ++j) {
				MSKint32t *iindices = (MSKint32t *) malloc(
						nonzeroes * sizeof(MSKint32t));
				MSKint32t *jindices = (MSKint32t *) malloc(
						nonzeroes * sizeof(MSKint32t));
				MSKrealt *values = (MSKrealt *) malloc(
						nonzeroes * sizeof(MSKrealt));
				iindices[0] = i;
				jindices[0] = j;
				values[0] = 1;
				MSKint64t idx;
				if (res == MSK_RES_OK)
					res = MSK_appendsparsesymmat(task, dim, nonzeroes, iindices,
							jindices, values, &idx);
				assert(runner < values_in_lower_triangle);
				running_ids[runner++] = idx;
				if (i != j) {
					for (int b = 0; b < numCuts && res == MSK_RES_OK; ++b) {
						res = MSK_putbaraij(task, constraint_number, b, 1, &idx,
								&weight5);
					}
					if (res == MSK_RES_OK)
						res = MSK_putconbound(task, constraint_number++,
								MSK_BK_UP, -numCuts, numCuts - 1.0);
				} else {
					for (MSKint64t b = 0; b < numCuts && res == MSK_RES_OK;
							++b) {
						res = MSK_putbaraij(task, constraint_number, b, 1, &idx,
								&weight1);
						if (res == MSK_RES_OK)
							res = MSK_putconbound(task, constraint_number++,
									MSK_BK_FX, 1.0, 1.0);
					}
				}
			}
		}

		declare_sumzero_constraint(&constraint_number);

		declare_branching_constraints(&constraint_number);


	}


	void setup_avg_cost() {
		if (res == MSK_RES_OK)
			res = MSK_putobjsense(task, MSK_OBJECTIVE_SENSE_MINIMIZE);

		cost_matrix.resize(numCuts);
		compute_cost_matrix();
	}

	void setup_max_cost() {
		cost_matrix.resize(numCuts);
		for (int b = 0; b < numCuts; ++b) {
			MSKrealt *cost_weights = (MSKrealt*) malloc(
					values_in_lower_triangle * sizeof(MSKrealt));
			get_cost_weights(b, cost_weights);
			res = MSK_putbaraij(task, constraint_number, b,
					values_in_lower_triangle, running_ids, cost_weights);
			res = MSK_putaij(task, constraint_number, z_id, -1);
			cost_constraint[b] = constraint_number;
			res = MSK_putconbound(task, constraint_number++, MSK_BK_UP, 0, 0);
		}
		if (res == MSK_RES_OK)
			res = MSK_putcj(task, z_id, 1);
	}

	void malloc_solution_container() {
		skc = (MSKstakeye *) malloc(numConstraints * sizeof(MSKstakeye));
		skx = (MSKstakeye *) malloc(numConstraints * sizeof(MSKstakeye));
		skn = (MSKstakeye *) malloc(numConstraints * sizeof(MSKstakeye));
		xc = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
		xx = (MSKrealt *) malloc(dim * dim * numCuts * sizeof(MSKrealt));
		y = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
		slc = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
		suc = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
		slx = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
		sux = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
		snx = (MSKrealt *) malloc(numConstraints * sizeof(MSKrealt));
	}

	void declare_sumzero_constraint(int* constraint_number) {
		MSKrealt *weights = (MSKrealt *) malloc(
				values_in_lower_triangle * sizeof(MSKrealt));
		for (int k = 0; k < values_in_lower_triangle; ++k) {
			weights[k] = 1.0;
		}
		for (int b = 0; b < numCuts && res == MSK_RES_OK; ++b) {
			res = MSK_putbaraij(task, *constraint_number, b,
					values_in_lower_triangle, running_ids, weights);
		}
		if (res == MSK_RES_OK)
			res = MSK_putconbound(task, (*constraint_number)++, MSK_BK_FX, 0.0,
					0.0);

		for (int b = 0; b < numCuts && res == MSK_RES_OK; ++b) {
			for (int i = 0; i < dim && res == MSK_RES_OK; ++i) {
				res = MSK_putaij(task, *constraint_number, IJtoK(b, i), 1.0);
			}
			if (res == MSK_RES_OK)
				res = MSK_putconbound(task, (*constraint_number)++, MSK_BK_FX,
						0.0, 0.0);
		}
	}

	void declare_branching_constraints(int* constraint_number) {
		const double weight5 = 0.5;
		const double weight1 = 1.0;
		branch_constraints = (MSKint64t *) malloc(
				dim * dim * sizeof(MSKint64t));
		for (int b = 0; b < numCuts; ++b)
			for (int i = 0; i < dim && res == MSK_RES_OK; ++i) {
				int indexI = max(i, (signed) referenceI);
				int indexJ = min(i, (signed) referenceI);
				int index = IJtoIndex(indexI, indexJ);
				assert(index < values_in_lower_triangle);
				MSKint64t idx = running_ids[index];
				if (indexI == indexJ)
					res = MSK_putbaraij(task, *constraint_number, b, 1, &idx,
							&weight1);
				else
					res = MSK_putbaraij(task, *constraint_number, b, 1, &idx,
							&weight5);
				if (res == MSK_RES_OK)
					res = MSK_putaij(task, *constraint_number, IJtoK(b, i),
							-referenceBranch[b]);
				if (res == MSK_RES_OK)
					res = MSK_putconbound(task, *constraint_number, MSK_BK_FX,
							0, 0);
				index = IJtoK(i, b);
				assert(index < dim * dim);
				branch_constraints[index] = (*constraint_number)++;
			}
	}

	void write_problem(std::string path) {
		MSK_writetask(task, (path + ".task").c_str());
	}

	void solve() {
		res = MSK_optimize(task);
		stringstream mskpath;
		mskpath << "model" << solverun << ".task";
		fallback_ = false;
		if (res == MSK_RES_TRM_STALL)
			res = MSK_RES_OK;
		if (res != MSK_RES_OK) {
			char symname[MSK_MAX_STR_LEN];
			char desc[MSK_MAX_STR_LEN];
			printf("An error occurred while optimizing.\n");
			MSK_getcodedesc(res, symname, desc);
			printf("Error %s - '%s'\n", symname, desc);
			exit(1);
		}

		res = MSK_getsolsta(task, MSK_SOL_ITR, &solsta);
		if (res != MSK_RES_OK)
			std::cout << res << std::endl;

		switch (solsta) {
		case MSK_SOL_STA_UNKNOWN: {
			std::cout << "Mosek reports a numerically unstable status. Please solve this test instance with our Gurobi framework." << std::endl;
			exit(1);
		}
		case MSK_SOL_STA_OPTIMAL:
		case MSK_SOL_STA_NEAR_OPTIMAL: {
			feasible = true;


			res = MSK_getdualobj(task, MSK_SOL_ITR, &objective_);

			break;
		}
		case MSK_SOL_STA_DUAL_INFEAS_CER:
		case MSK_SOL_STA_PRIM_INFEAS_CER:
		case MSK_SOL_STA_NEAR_DUAL_INFEAS_CER:
		case MSK_SOL_STA_NEAR_PRIM_INFEAS_CER:
			feasible = false;
			break;
		default:
			std::cerr << "Unknown or weird solution status. Giving up. "
					<< solsta << std::endl;
			exit(1);
		}

	}

	double getLinearPrimalValue(int i, int b) {
		if (res == MSK_RES_OK)
			res = MSK_getxx(task, MSK_SOL_ITR, xx);
		if (res != MSK_RES_OK) {
			cout << "Unable to fetch linera primal variable " << b << ", " << i
					<< endl;
			exit(1);
		}
		return xx[IJtoK(b, i)];
	}

	double getSdpPrimalvalue(int i, int j, int b) {
		if (res == MSK_RES_OK)
			res = MSK_getbarxj(task, MSK_SOL_ITR, b, Y);
		if (res != MSK_RES_OK) {
			cout << "Unable to fetch sdp primal variable " << b << ", " << i
					<< endl;
			exit(1);
		}
		if (i >= j)
			return Y[ijToMosekIndex(i, j)];
		else
			return Y[ijToMosekIndex(j, i)];
	}

	double getSdpDualvalue(int constraint) {
		return y[constraint];
	}

	double getSdpBvalue(int i) {
		MSKboundkeye bk;
		MSKrealt bl;
		MSKrealt bu;
		if (res == MSK_RES_OK)
			res = MSK_getconbound(task, i, &bk, &bl, &bu);
		if (res != MSK_RES_OK) {
			cout << "Unable to fetch rhs " << i << endl;
			exit(1);
		}
		switch (bk) {
		case (MSK_BK_LO):
		case (MSK_BK_FX):
			return bl;
		case (MSK_BK_UP):
			return bu;
		default:
			return 0.0;
		}
	}


	int ijToMosekIndex(int i, int j) {
		return i + j * dim - j * (j + 1) / 2;
	}

	double getSdpCvalue(int i, int j, int b) {
		int ii = i > j ? i : j;
		int jj = i > j ? j : i;

		return cost_matrix[b][ii * (ii + 1) / 2 + jj];
	}

	void writeBounds(int i, int j, int b, double lhs, double rhs) {
		return;


		int branchKey = (unsigned) i == referenceI ? j : i;
		if (abs(rhs) < eps_integrality) {
			res = MSK_putvarbound(task, IJtoK(b - 1, branchKey - 1), MSK_BK_RA,
					-1.0, 1.0);
		} else {
			double bound = rhs * referenceBranch[b];
			res = MSK_putvarbound(task, IJtoK(b - 1, branchKey - 1), MSK_BK_FX,
					bound, bound);
		}
		if (res != MSK_RES_OK) {
			std::cerr << "Unable to branch on variable " << branchKey << ", "
					<< b << std::endl;
			exit(1);
		}
	}

	void setLinVarBound(int i, int b, double lb, double ub) {
		if(i == -1 && b == 0){
			done = true;
			return;
		}
		MSKboundkeye key = ((lb == ub) ? MSK_BK_FX : MSK_BK_RA);
		res = MSK_putvarbound(task, IJtoK(b, i), key, lb, ub);
		if (res != MSK_RES_OK) {
			std::cerr << "Unable to branch on variable " << i << ", " << b
					<< std::endl;
			exit(1);
		}
	}

	double get_variable_fixation(int i, int b) {
		MSKboundkey_enum bk;
		double lb;
		double ub;
		MSK_getvarbound(task, IJtoK(b, i), &bk, &lb, &ub);
		if (bk != MSK_BK_FX)
			return 0.0;
		return ub;
	}

	void compute_cost_matrix() {
		for (int b = 0; b < numCuts && res == MSK_RES_OK; ++b) {
			MSKrealt *cost_weights = (MSKrealt*) malloc(
					values_in_lower_triangle * sizeof(MSKrealt));
			get_cost_weights(b, cost_weights);
			res = MSK_putbarcj(task, b, values_in_lower_triangle, running_ids,
					cost_weights);
			for (int i = 0; i < dim && res == MSK_RES_OK; ++i) {
				res = MSK_putcj(task, IJtoK(b, i), linearcost(b, i));
			}
		}
		if (res == MSK_RES_OK)
			res = MSK_putcfix(task, constantcost());
		assert(res == MSK_RES_OK);
	}

	void compute_cost_constraint() {
		for (int b = 0; b < numCuts && res == MSK_RES_OK; ++b) {
			MSKrealt *cost_weights = (MSKrealt*) malloc(
					values_in_lower_triangle * sizeof(MSKrealt));
			get_cost_weights(b, cost_weights);
			res = MSK_putbaraij(task, b, cost_constraint[b],
					values_in_lower_triangle, running_ids, cost_weights);
			for (int i = 0; i < dim && res == MSK_RES_OK; ++i) {
				res = MSK_putaij(task, cost_constraint[b], IJtoK(b, i),
						linearcost(b, i));
			}
		}
		if (res == MSK_RES_OK)
			res = MSK_putcfix(task, constantcost());
	}

	void get_cost_weights(int b, double *cost_weights) {
		int runner = 0;
		for (int i = 0; i < dim; ++i)
			for (int j = 0; j <= i; ++j) {
				double value = cost(b, i, j);
				cost_weights[runner++] = value;
			}
		cost_matrix[b] = cost_weights;
	}

	bool lp_infeasible() {
		return !feasible;
	}

	bool lp_feasible() {
		return feasible;
	}

	int KtoI(int k) {
		return (k / dim);
	}

	int KtoJ(int k) {
		return (k % dim);
	}

	int IJtoK(int i, int j) {
		return i * dim + j;
	}

	int IJtoIndex(int i, int j) {
		if (i >= j)
			return i * (i + 1) / 2 + j;
		else
			return j * (j + 1) / 2 + i;
	}

}
;

#endif /* MOSEKSOLVER_H_ */
