#ifndef _SDPSOLVER_H_
#define _SDPSOLVER_H_

#include <math.h>
#include <stdlib.h>
#include <vector>
#include <random>
#include "matutils.h"
#include <bitset>
#include <fstream>
#include <iostream>
#include <algorithm>

using namespace std;

struct distancePointTuple {
	double distance;
	int indexI;
	int indexJ;

	distancePointTuple(double d, int i, int j) :
			distance(d), indexI(i), indexJ(j) {
	}

};

inline bool comp_function(distancePointTuple a, distancePointTuple b) {
	return a.distance < b.distance;
}

class sdpSolver {

public:

	int id;

	/*
	 * The problem and solution data.
	 */

	vector<int> referenceBranch;
	unsigned int referenceI = 0;
	unsigned int referenceSlot = 0;

	/*
	 * Storage for the initial and final solutions.
	 */

	double pobj, dobj;

	const double eps_; 

	int dim;
	int dimNoDummies;
	int numBlocks;
	int numConstraints;
	int numCuts;
	int totalYdimension;
	double normalization;
	double objective_;
	int candidate_;
	int depth;
	std::vector<Matrix> c_outer;
	std::vector<Matrix> c_zeroone;
	std::vector<Matrix> c_onezero;
	std::vector<Matrix> c_oneone;
	std::vector<Matrix> c_zerozero;
	Matrix probabilities;
	Matrix distances;
	double eps_integrality;
	double const_offset;
	const bool max_objective = false;
	bool swapDandP = false;
	const double eps_sdp_property = 0.001;
	bool fallback_;
	int solverun = 0;
	const double minOverEmpty = 0; 
	bool done = false;
	int numfixvar;


	vector<vector<distancePointTuple>> cut_outer_distances;
	vector<vector<distancePointTuple>> cut_zeroone_distances;
	vector<vector<distancePointTuple>> cut_onezero_distances;
	vector<vector<distancePointTuple>> cut_oneone_distances;
	vector<vector<distancePointTuple>> cut_zerozero_distances;
	vector<vector<distancePointTuple>> diagonal_one_distances;
	vector<vector<distancePointTuple>> diagonal_zero_distances;

	typedef int Basis;

	sdpSolver() :
			id(0), referenceI(0), pobj(0.0), dobj(0.0), eps_(1e-6), dim(0), dimNoDummies(
					0), numBlocks(0), numConstraints(0), numCuts(0), totalYdimension(
					0), normalization(0.0), objective_(0), candidate_(0), depth(
					0), eps_integrality(0.0001), const_offset(0.0), fallback_(
					false) {
	}

	virtual ~sdpSolver() {

	}

	void read(char* filename, char* format) {
		ifstream fin(filename);

		int nn;
		ignoreComment(fin);
		fin >> nn;
		dimNoDummies = nn;

		numCuts = (int) ceil(log2(nn));
		dim = (int) (pow(2, numCuts));

		probabilities = emptyMatrix(dimNoDummies);
		distances = emptyMatrix(dimNoDummies);

		ignoreComment(fin);
		if (swapDandP) {
			readMatrix(distances, fin);
			ignoreComment(fin);
			readMatrix(probabilities, fin);
		} else {
			readMatrix(probabilities, fin);
			ignoreComment(fin);
			readMatrix(distances, fin);
		}

		ignoreComment(fin);
		vector<LetterSlot> fixations = readFixations(fin);

		bool fixed = false;

		for (auto const& ls : fixations) {
			fixed = true;
			referenceI = ls.letter;
			referenceSlot = ls.slot;
		}

		if (!fixed) {
			referenceI = dimNoDummies;
			referenceSlot = referenceI;
			if (dimNoDummies == dim) {
				dim = dim * 2;
				++numCuts;
			}
		}

		ignoreComment(fin);
		scale(probabilities, fin);

		if (checkSymmetry(distances)) {
			symmetrizeMatrix(probabilities);
		} else if (checkSymmetry(probabilities)) {
			symmetrizeMatrix(distances);
		}

		initCost(probabilities, distances);

		referenceBranch.resize(numCuts);

		std::bitset<sizeof(long)> bits(referenceSlot);
		for (int b = 0; b < numCuts; ++b) {
			bool bitSet = bits.test(b);
			int branchValue = bitSet ? 1 : -1;
			referenceBranch[b] = branchValue;
		}


		double trivialBound = computeTrivialBound();
		std::cout << "Trivial bound: " << trivialBound << std::endl;

		numConstraints = 2 * numCuts * dim + dim * (dim - 1) / 2 + numCuts + 1;
		
		if (max_objective)
			numConstraints += numCuts;
		totalYdimension = dim * numCuts + dim * dim; 
		if (max_objective)
			totalYdimension += numCuts + 1; 
		numBlocks = numCuts + dim;
		if (max_objective)
			numBlocks += 1;		

		normalization = 1.0 / numCuts;
		const_offset = 0;

		init_env();

		if (max_objective)
			add_max_variable();

		declare_constraints();

		if (max_objective)
			setup_max_cost();
		else
			setup_avg_cost();

		checkSanityOfSetup();

		for (auto const& ls : fixations) {
			for (int b = 0; b < numCuts; ++b) {
				double value = checkBit(ls.slot, b) ? 1.0 : -1.0;
				init_bounds(b, ls.letter, value);
			}
		}

		for (int i = dimNoDummies; i < dim; ++i) {
			for (int b = 0; b < numCuts; ++b) {
				double value = checkBit(i, b) ? 1.0 : -1.0;
				init_bounds(b, i, value);
			}
		}
	}

	void addConstToMatrix(Matrix* M, int size, double c) {
		for (int i = 0; i < size; ++i)
			for (int j = 0; j < size; ++j)
				(*M)[i][j] += c;
	}

	double computeTrivialBound() {
		double min_dist = find_min_in_matrix(distances);
		double bound = 0.0;
		for (int i = 0; i < dimNoDummies; ++i)
			for (int j = 0; j < dimNoDummies; ++j)
				bound += probabilities[i][j];
		bound = bound * min_dist;
		return bound;
	}

	void shuffleMatrix(Matrix& M, vector<int> permutation) {
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j < dimNoDummies; ++j) {
				M[i][j] = M[permutation[i]][permutation[j]];
			}
		}
		computeCostFromScratch();
	}

	void shuffleDistances(vector<int> permutation) {
		shuffleMatrix(distances, permutation);
	}

	void shuffleProbabilities(vector<int> permutation) {
		shuffleMatrix(probabilities, permutation);
	}

	bool checkSymmetry(Matrix& M) {
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j < i; ++j) {
				if (M[i][j] != M[j][i])
					return false;
			}
		}
		return true;
	}

	void symmetrizeMatrix(Matrix& M) {
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j <= i; ++j) {
				M[i][j] = 0.5 * (M[i][j] + M[j][i]);
				M[j][i] = M[i][j];
			}
		}
	}

	virtual void init_env() = 0;

	virtual void add_max_variable() = 0;

	virtual void declare_constraints() = 0;

	virtual void setup_max_cost() = 0;

	virtual void setup_avg_cost() = 0;

	virtual void checkSanityOfSetup() = 0;

	virtual void solve() = 0;

	virtual void write_problem(std::string path) = 0;

	void init_cut_distances() {
		cut_outer_distances.resize(numCuts);
		cut_zeroone_distances.resize(numCuts);
		cut_onezero_distances.resize(numCuts);
		cut_oneone_distances.resize(numCuts);
		cut_zerozero_distances.resize(numCuts);
		diagonal_one_distances.resize(numCuts);
		diagonal_zero_distances.resize(numCuts);
		for (int b = 0; b < numCuts; ++b) {
			for (int k = 0; k < dimNoDummies; ++k) {
				for (int l = 0; l < dimNoDummies; ++l) {
					if (k == l) {
						if (checkBit(k, b)) {
							diagonal_one_distances[b].push_back(
									distancePointTuple(distances[k][k], k, k));
						} else {
							diagonal_zero_distances[b].push_back(
									distancePointTuple(distances[k][k], k, k));
						}
						continue;
					}
					if (!checkBit(k, b) && checkBit(l, b)) {
						cut_zeroone_distances[b].push_back(
								distancePointTuple(distances[k][l], k, l));
						cut_outer_distances[b].push_back(
								distancePointTuple(distances[k][l], k, l));
					}
					else if (checkBit(k, b) && !checkBit(l, b)) {
						cut_onezero_distances[b].push_back(
								distancePointTuple(distances[k][l], k, l));
						cut_outer_distances[b].push_back(
								distancePointTuple(distances[k][l], k, l));
					}
					else if (checkBit(k, b) && checkBit(l, b)) {
						cut_oneone_distances[b].push_back(
								distancePointTuple(distances[k][l], k, l));
					}
					else if (!checkBit(k, b) && !checkBit(l, b)) {
						cut_zerozero_distances[b].push_back(
								distancePointTuple(distances[k][l], k, l));
					}
				}
			}
			std::sort(cut_outer_distances[b].begin(),
					cut_outer_distances[b].end(), comp_function);
			std::sort(cut_zeroone_distances[b].begin(),
					cut_zeroone_distances[b].end(), comp_function);
			std::sort(cut_onezero_distances[b].begin(),
					cut_onezero_distances[b].end(), comp_function);
			std::sort(cut_oneone_distances[b].begin(),
					cut_oneone_distances[b].end(), comp_function);
			std::sort(cut_zerozero_distances[b].begin(),
					cut_zerozero_distances[b].end(), comp_function);
			std::sort(diagonal_one_distances[b].begin(),
					diagonal_one_distances[b].end(), comp_function);
			std::sort(diagonal_zero_distances[b].begin(),
					diagonal_zero_distances[b].end(), comp_function);
		}
	}

	void assertSymmetricAppearance(vector<vector<distancePointTuple>>& list) {
		for (int b = 0; b < numCuts; ++b) {
			for (vector<distancePointTuple>::iterator it = list[b].begin();
					it != list[b].end(); ++it) {
				int k = it->indexI;
				int l = it->indexJ;
				double distance = it->distance;
				bool found = false;
				double checkDist = -1;
				for (vector<distancePointTuple>::iterator it2 = list[b].begin();
						it2 != list[b].end(); ++it2) {
					int correspL = it2->indexI;
					int correspK = it2->indexJ;
					if (correspL == l && correspK == k) {
						found = true;
						checkDist = it2->distance;
					}
				}
				assert(found);
				assert(distance == checkDist);
			}
		}
	}

	double orig_cost(int i, int j, int k, int l) {
		if (i >= dimNoDummies || j >= dimNoDummies || k >= dimNoDummies
				|| l >= dimNoDummies)
			return 0;
		return probabilities[i][j] * distances[k][l];
	}

	double cost(const unsigned int& b, const unsigned int& i,
			const unsigned int& j) {
		if (i >= (unsigned) dimNoDummies || j >= (unsigned) dimNoDummies)
			return 0.0;
		double res =
				i == j ?
						0.0 :
						0.125
								* ((c_oneone[b][i][j] - c_onezero[b][i][j]
										- c_zeroone[b][i][j]
										+ c_zerozero[b][i][j])
										* probabilities[i][j]
										+ (c_oneone[b][j][i]
												- c_onezero[b][j][i]
												- c_zeroone[b][j][i]
												+ c_zerozero[b][j][i])
												* probabilities[j][i]);



		return normalization * res;
	}

	double constantcost() {
		double res = 0.0;
		for (int b = 0; b < numCuts; ++b)
			for (int i = 0; i < dimNoDummies; ++i)
				for (int k = 0; k < dimNoDummies; ++k) {
					if (i == k) {
						res += 0.5 * probabilities[i][i]
								* (c_oneone[b][i][i] + c_zerozero[b][i][i]);
					} else {
						res += 0.25
								* (c_zeroone[b][i][k] + c_onezero[b][i][k]
										+ c_oneone[b][i][k]
										+ c_zerozero[b][i][k])
								* probabilities[i][k];
					}

				}
		return normalization * res + const_offset;
	}

	double linearcost(int b, int i) {
		if (i >= dimNoDummies)
			return 0.0;
		double res = 0.0;
		for (int k = 0; k < dimNoDummies; ++k) {
			res += 0.25
					* (probabilities[i][k]
							* (c_oneone[b][i][k] + c_onezero[b][i][k]
									- c_zeroone[b][i][k] - c_zerozero[b][i][k])
							+ probabilities[k][i]
									* (c_oneone[b][k][i] - c_onezero[b][k][i]
											+ c_zeroone[b][k][i]
											- c_zerozero[b][k][i]));
		}

		return normalization * res;
	}

	bool checkBit(int var, int pos) {
		return ((var) & (1 << (pos)));
	}

	double find_max_in_matrix(Matrix M) {
		double max = -1;
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j < dimNoDummies; ++j) {
				if (M[i][j] > max)
					max = M[i][j];
			}
		}
		return max;
	}

	double find_min_in_matrix(Matrix M) {
		double min = -1;
		bool firstIter = true;
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j < dimNoDummies; ++j) {
				if (firstIter || M[i][j] < min) {
					firstIter = false;
					min = M[i][j];
				}
			}
		}
		return min;
	}

	void initCost(Matrix P, Matrix D) {
		init_cut_distances();
		c_outer.resize(numCuts);
		c_zeroone.resize(numCuts);
		c_onezero.resize(numCuts);
		c_oneone.resize(numCuts);
		c_zerozero.resize(numCuts);
		for (int cut = 0; cut < numCuts; ++cut) {
			c_outer[cut] = emptyMatrix(dim);
			c_zeroone[cut] = emptyMatrix(dim);
			c_onezero[cut] = emptyMatrix(dim);
			c_oneone[cut] = emptyMatrix(dim);
			c_zerozero[cut] = emptyMatrix(dim);
			for (int i = 0; i < dim; ++i) {
				for (int j = 0; j < dim; ++j) {
					if (i >= dimNoDummies || j >= dimNoDummies) {
						c_outer[cut][i][j] = 0.0;
						c_zeroone[cut][i][j] = 0.0;
						c_onezero[cut][i][j] = 0.0;
						c_oneone[cut][i][j] = 0.0;
						c_zerozero[cut][i][j] = 0.0;
						continue;
					}
					if (i == j) {
						c_outer[cut][i][j] = 0.0;
						c_zeroone[cut][i][j] = minOverEmpty;
						c_onezero[cut][i][j] = minOverEmpty;
						c_oneone[cut][i][j] = nullOrDistance(
								diagonal_one_distances[cut]);
						c_zerozero[cut][i][j] = nullOrDistance(
								diagonal_zero_distances[cut]);
					} else {
						c_zeroone[cut][i][j] = nullOrDistance(
								cut_outer_distances[cut]);
						c_zeroone[cut][i][j] = nullOrDistance(
								cut_zeroone_distances[cut]);
						c_onezero[cut][i][j] = nullOrDistance(
								cut_onezero_distances[cut]);
						c_oneone[cut][i][j] = nullOrDistance(
								cut_oneone_distances[cut]);
						c_zerozero[cut][i][j] = nullOrDistance(
								cut_zerozero_distances[cut]);
					}
				}
			}
		}
	}

	double nullOrDistance(vector<distancePointTuple>& t) {
		if (t.empty())
			return minOverEmpty;
		return t.front().distance;
	}

	void computeCostFromScratch() {
		initCost(probabilities, distances);
		compute_cost_matrix();
	}

	bool update_costs(int i, bool addedConstraints) {
		if (i >= dimNoDummies)
			return false;
		bool changed = false;
		for (int b = 0; b < numCuts; ++b) {
			for (int j = 0; j < dimNoDummies; ++j) {
				if (i == j) {
					update_cost_vector(c_oneone, diagonal_one_distances, i, j,
							b, &changed, "one-one", addedConstraints, true,
							true);
					update_cost_vector(c_zerozero, diagonal_zero_distances, i,
							j, b, &changed, "zero-zero", addedConstraints,
							false, false);
				} else {
					symmetric_update_cost_vector(c_outer, cut_outer_distances,
							i, j, b, &changed, "outer");
					update_cost_vector(c_zeroone, cut_zeroone_distances, i, j,
							b, &changed, "zero-one", addedConstraints, false,
							true);
					update_cost_vector(c_onezero, cut_onezero_distances, i, j,
							b, &changed, "one-zero", addedConstraints, true,
							false);
					update_cost_vector(c_oneone, cut_oneone_distances, i, j, b,
							&changed, "one-one", addedConstraints, true, true);
					update_cost_vector(c_zerozero, cut_zerozero_distances, i, j,
							b, &changed, "zero-zero", addedConstraints, false,
							false);
				}
			}
		}
		return changed;
	}

	void symmetric_update_cost_vector(std::vector<Matrix>& vec_to_change,
			vector<vector<distancePointTuple>>& sorted_distances, int i, int j,
			int b, bool *changed, std::string description) {
		for (vector<distancePointTuple>::iterator it =
				sorted_distances[b].begin(); it != sorted_distances[b].end();
				++it) {
			int k = it->indexI;
			int l = it->indexJ;
			if ((placement_possible(i, k) && placement_possible(j, l))
					|| (placement_possible(i, l) && placement_possible(j, k))) {
				double old_cost = vec_to_change[b][i][j];
				vec_to_change[b][i][j] = it->distance;
				vec_to_change[b][j][i] = vec_to_change[b][i][j];
				bool this_changed = old_cost != vec_to_change[b][i][j];
				*changed = *changed || this_changed;
				break;
			}
		}
	}

	bool update_cost_vector(std::vector<Matrix>& vec_to_change,
			const vector<vector<distancePointTuple>>& sorted_distances, int i,
			int j, int b, bool *changed, std::string description,
			bool addedConstraints, bool alpha, bool beta) {
		for (vector<distancePointTuple>::const_iterator it =
				sorted_distances[b].begin(); it != sorted_distances[b].end();
				++it) {
			int k = it->indexI;
			int l = it->indexJ;
			if ((placement_possible(i, k) && placement_possible(j, l))) {
				double old_cost = vec_to_change[b][i][j];
				vec_to_change[b][i][j] = it->distance;
				bool this_changed = old_cost != vec_to_change[b][i][j];
				*changed = *changed || this_changed;
				return true;
			}
		}
		double old_cost = vec_to_change[b][i][j];
		vec_to_change[b][i][j] = minOverEmpty;
		bool this_changed = old_cost != vec_to_change[b][i][j];
		*changed = *changed || this_changed;
		return false;
	}

	bool placement_possible(int i, int k) {
		if ((unsigned) k == referenceSlot) {
			return (unsigned) i == referenceI;
		}
		if ((unsigned) i == referenceI) {
			return (unsigned) k == referenceSlot;
		}
		for (int b = 0; b < numCuts; ++b) {
			double y_i_b = get_variable_fixation(i, b);
			if (abs(y_i_b) > eps_) {
				bool value = abs((y_i_b + 1.0) / 2.0) > eps_integrality;
				if (value != checkBit(k, b)) {
					return false;
				}
			}
		}
		return true;
	}

	bool variable_is_fixed(int i, int b) {
		return abs(get_variable_fixation(i, b)) > eps_integrality;
	}

	double QAP_objective() {

		std::vector<int> orig_X = recoverX();
		return QAP_objective(orig_X);

	}

	double QAP_objective(std::vector<int> permutation) {
		double sum = 0.0;

		bool bits[dimNoDummies];
		for (int i = 0; i < dimNoDummies; ++i)
			bits[i] = false;
		for (int i = 0; i < dimNoDummies; ++i) {
			bits[permutation[i]] = true;
		}
		for (int i = 0; i < dimNoDummies; ++i) {
			if (!bits[i]) {
				std::cout << "Illegal permutation!" << std::endl;
				exit(1);
			}
		}

		for (int i = 0; i < dimNoDummies; ++i) {
			for (int j = 0; j < dimNoDummies; ++j) {
				sum += probabilities[i][j]
						* distances[permutation[i]][permutation[j]];
			}
		}

		return sum;
	}

	double abs(double n) {
		return n >= 0 ? n : -n;
	}

	std::vector<int> recoverX() {
		std::vector<int> orig_X(dimNoDummies);

		for (int i = 0; i < dimNoDummies; ++i) {
			std::bitset<sizeof(long)> bits;
			bits.reset();
			for (int b = 0; b < numCuts; ++b) {
				double val = getLinearPrimalValue(i, b);
				if (abs(val - 1) <= eps_integrality) {
					bits.set(b, true);
				} else {
					assert(abs(val + 1) <= eps_integrality); 
					bits.set(b, false);
				}
			}
			orig_X[i] = bits.to_ulong();
		}

		if (referenceI < (unsigned) dimNoDummies) {
			orig_X[referenceI] = referenceSlot;
		}
		return orig_X;
	}

	void print_objective_per_cut() {
		for (int b = 0; b < numCuts; ++b) {
			double sum = 0.0;
			for (int i = 0; i < dim; ++i)
				for (int j = 0; j < dim; ++j) {
					sum += getSdpCvalue(i, j, b) * getSdpPrimalvalue(i, j, b);
				}
			std::cout << sum << std::endl;
		}
	}

	virtual double getSdpPrimalvalue(int i, int j, int b) = 0;

	virtual double getLinearPrimalValue(int i, int b) = 0;

	virtual double getSdpDualvalue(int constraint) = 0;

	virtual double getSdpBvalue(int i) = 0;

	virtual double getSdpCvalue(int i, int j, int b) = 0;

	void init_bounds(int b, int i, double value) {
		set_bounds(biToVariable(b, i), value, value);
	}

	void set_bounds(const unsigned int& variable, const double& lowerBound,
			const double& upperBound) {

		unsigned int bb = variableToB(variable);
		unsigned int i = variableToI(variable);
		if (i < 0)
			std::cerr << "index too small: " << i << std::endl;
		setLinVarBound(i, bb, lowerBound, upperBound);

		unsigned int indexI = i < referenceI ? i : referenceI;
		unsigned int indexJ = i < referenceI ? referenceI : i;
		double lhs = 0.0;
		double rhs = 0.0;
		double lb = lowerBound;
		double ub = upperBound;
		if (lowerBound > -1 + 1e-6)
			lb = 1;
		if (upperBound < 1 - 1e-6)
			ub = -1;
		if (lb == ub) { 
			++depth;
			lhs = 0.5;
			if (lb == referenceBranch[bb]) { 
				rhs = 1;
			} else { 
				rhs = -1;
			}
		} else { 
			lhs = 0.0;
			--depth;
		}

		writeBounds(indexI, indexJ, bb, lhs, rhs);

		update_all_costs(lb == ub);
	}

	virtual void setLinVarBound(int i, int b, double lb, double ub) = 0;

	bool update_all_costs(bool addedConstraints) {
		bool res = false;
		for (int i = 0; i < dimNoDummies; ++i)
			res = update_costs(i, addedConstraints) || res;
		max_objective ? compute_cost_constraint() : compute_cost_matrix();
		return res;
	}

	void writeCost(const char * path) {
		ofstream out;
		out.open(path);
		for (int b = 0; b < numCuts; ++b) {
			for (int i = 0; i < dimNoDummies; ++i) {
				for (int j = 0; j < dimNoDummies; ++j) {
					out << cost(b, i, j) << " ";
				}
				out << endl;
			}
			out << endl;
		}
		for (int b = 0; b < numCuts; ++b) {
			for (int i = 0; i < dimNoDummies; ++i) {
				out << linearcost(b, i) << " ";
			}
			out << endl;
		}
		out << endl << constantcost() << endl;
		out.close();
	}

	void printTotalEntrySum() {
		double sum = 0.0;
		for (int b = 0; b < numCuts; ++b) {
			double sumb = 0.0;
			for (int i = 0; i < dim; ++i)
				for (int j = 0; j < dim; ++j) {
					sumb += getSdpPrimalvalue(i, j, b);
				}
			std::cout << "Entry sum for Y^" << b << ": " << sumb << std::endl;
			sum += sumb;
		}
		std::cout << "Total entry sum for Y: " << sum << std::endl;
	}


	virtual void compute_cost_matrix() = 0;

	virtual void compute_cost_constraint() = 0;

	virtual double get_variable_fixation(int i, int b) = 0;

	virtual void writeBounds(int i, int j, int b, double lhs, double rhs) = 0;

	bool is_binary_variable(const int i) {
		return true;
	}

	virtual bool lp_infeasible() = 0;

	virtual bool lp_feasible() = 0;

	bool lp_integral() {
		return get_most_fractional_variable() == -1;
	}

	double objective() {
		return objective_;
	}

	double value_of(const unsigned int& variable) {
		int b = variableToB(variable);
		int i = variableToI(variable);
		return value_of_Y(i, b);
	}

	double value_of_Y(int i, int b) {
		return getSdpPrimalvalue(i, referenceI, b) * referenceBranch[b];
	}

	double lower_bound_of(const unsigned int& variable) {
		int i = variableToI(variable);
		int b = variableToB(variable);
		double value = get_variable_fixation(i, b);
		if (value < 1)
			value = -1;
		return value;
	}

	double upper_bound_of(const unsigned int& variable) {
		int i = variableToI(variable);
		int b = variableToB(variable);
		double value = get_variable_fixation(i, b);
		if (value > -1)
			value = 1;
		return value;
	}

	double frac(const double& f) {
		return f - trunc(f);
	}

	int get_candidate() {

		int cand = get_most_fractional_variable();
		if (cand == -1)
			return get_arbitrary_unset_variable();
		return cand;
	}

	int get_most_fractional_variable() {
		int cand = -1;
		double minVal = 1 - eps_integrality;		
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int b = 0; b < numCuts; ++b) {
				double val = getLinearPrimalValue(i, b);
				if (val < 0)
					val = -val;
				if (val > 1 + eps_integrality) {
					std::cout << "Domain constraint violated: " << val
							<< std::endl;
				}
				if (val < minVal) {
					minVal = val;
					cand = biToVariable(b, i);
				}

			}
		}
		return cand;
	}

	int get_arbitrary_unset_variable() {
		for (int i = 0; i < dimNoDummies; ++i) {
			for (int b = 0; b < numCuts; ++b) {
				if (get_variable_fixation(i, b) == 0) {
					return biToVariable(b, i);
				}
			}
		}
		return -1;
	}

	

	double getPrimalCost() {
		double cost = 0.0;
		for (int b = 0; b < numCuts; ++b)
			for (int i = 0; i < dim; ++i)
				for (int j = 0; j < dim; ++j) {
					cost += getSdpCvalue(i, j, b) * getSdpPrimalvalue(i, j, b);
				}
		return cost;
	}

	double computeQuadCostInY() {
		double res = 0;
		for (int b = 0; b < numCuts; ++b) {
			for (int i = 0; i < dimNoDummies; ++i) {
				double y_ib = getLinearPrimalValue(i, b);
				for (int j = 0; j < dimNoDummies; ++j) {
					if (i == j) {
						res += probabilities[i][i]
								* (c_oneone[b][i][i] * (1.0 + y_ib) / 2.0
										+ c_zerozero[b][i][i] * (1.0 - y_ib)
												/ 2.0);
					} else {
						double y_jb = getLinearPrimalValue(j, b);
						res += 0.25 * probabilities[i][j]
								* (c_oneone[b][i][j] + c_onezero[b][i][j]
										+ c_zeroone[b][i][j]
										+ c_zerozero[b][i][j]
										+ (c_oneone[b][i][j]
												+ c_onezero[b][i][j]
												- c_zeroone[b][i][j]
												- c_zerozero[b][i][j]) * y_ib
										+ (c_oneone[b][i][j]
												- c_onezero[b][i][j]
												+ c_zeroone[b][i][j]
												- c_zerozero[b][i][j]) * y_jb
										+ (c_oneone[b][i][j]
												- c_onezero[b][i][j]
												- c_zeroone[b][i][j]
												+ c_zerozero[b][i][j]) * y_ib
												* y_jb);
					}
				}
			}
		}
		return normalization * res;
	}

	double computeRankRelaxationEffect() {
		double effect = 0;
		for (int b = 0; b < numCuts; ++b) {
			for (int i = 0; i < dimNoDummies; ++i) {
				double z_ib = getZib(i, b);
				for (int j = 0; j < dimNoDummies; ++j) {
					const double z_jb = getZib(j, b);
					const double Y_b_ij = getSdpPrimalvalue(i, j, b);
					const double y_b_i = getLinearPrimalValue(i, b);
					const double y_b_j = getLinearPrimalValue(j, b);

					const double z_b_ij_oneone =
							i == j ?
									(1 + y_b_i) / 2.0 :
									(1 + y_b_i + y_b_j + Y_b_ij) / 4.0;
					const double z_b_ij_zeroone =
							i == j ? 0.0 : (1 - y_b_i + y_b_j - Y_b_ij) / 4.0;
					const double z_b_ij_onezero =
							i == j ? 0.0 : (1 + y_b_i - y_b_j - Y_b_ij) / 4.0;
					const double z_b_ij_zerozero =
							i == j ?
									(1 - y_b_i) / 2.0 :
									(1 - y_b_i - y_b_j + Y_b_ij) / 4.0;

					const double oneone_discrepancy =
							i == j ?
									z_ib - z_b_ij_oneone :
									z_ib * z_jb - z_b_ij_oneone;
					const double zeroone_discrepancy =
							i == j ?
									0 - z_b_ij_zeroone :
									(1 - z_ib) * z_jb - z_b_ij_zeroone;
					const double onezero_discrepancy =
							i == j ?
									0 - z_b_ij_onezero :
									z_ib * (1 - z_jb) - z_b_ij_onezero;
					const double zerozero_discrepancy =
							i == j ?
									(1 - z_ib) - z_b_ij_zerozero :
									(1 - z_ib) * (1 - z_jb) - z_b_ij_zerozero;

					const double oneone_effect = probabilities[i][j]
							* c_oneone[b][i][j] * oneone_discrepancy;
					const double zeroone_effect = probabilities[i][j]
							* c_zeroone[b][i][j] * zeroone_discrepancy;
					const double onezero_effect = probabilities[i][j]
							* c_onezero[b][i][j] * onezero_discrepancy;
					const double zerozero_effect = probabilities[i][j]
							* c_zerozero[b][i][j] * zerozero_discrepancy;

					effect += oneone_effect + zeroone_effect + onezero_effect
							+ zerozero_effect;

				}
			}
		}
		return normalization * effect;
	}

	double computeQuadCostInZ() {
		double res = 0;
		for (int b = 0; b < numCuts; ++b) {
			for (int i = 0; i < dimNoDummies; ++i) {
				double z_ib = getZib(i, b);
				for (int j = 0; j < dimNoDummies; ++j) {
					double z_jb = getZib(j, b);
					if (i == j) {
						res += probabilities[i][i]
								* (c_oneone[b][i][i] * z_ib
										+ c_zerozero[b][i][i] * (1.0 - z_ib));
					} else {
						res += probabilities[i][j]
								* (c_oneone[b][i][j] * z_ib * z_jb
										+ c_onezero[b][i][j] * z_ib * (1 - z_jb)
										+ c_zeroone[b][i][j] * (1 - z_ib) * z_jb
										+ c_zerozero[b][i][j] * (1 - z_ib)
												* (1 - z_jb));
					}
				}
			}
		}
		return normalization * res;
	}

	double getZib(int i, int b) {
		double y_ib = getLinearPrimalValue(i, b);
		return (y_ib + 1.0) / 2.0;
	}

	double getDualCost() {
		double cost = 0.0;
		for (int i = 0; i < numConstraints; ++i) {
			double rhs = getSdpBvalue(i);
			double yi = getSdpDualvalue(i);
			cost += rhs * yi;
		}
		return cost;
	}

	int variableToI(int variable) {
		return (variable % dim);
	}

	int variableToB(int variable) {
		return variable / dim;
	}

	int biToVariable(int cut, int i) {
		return cut * dim + i;
	}

	void setID(int i) {
		id = i;
	}

	int getID() {
		return id;
	}

	int getDim() {
		return dim;
	}

	bool fallback() {
		return fallback_;
	}

}
;

#endif
